void bind_ejecta_distribution(double *radial_distance, double *initial_bearing, int num_elements, double crater_diameter, double ejrim, double ejecta_truncation, int dorays, double *ejecta_thickness);
void bind_ejecta_profile(double *radial_distance, int num_elements, double crater_diameter, double ejrim, double *elevation);
void bind_ejecta_ray_intensity(double *radial_distance, double *initial_bearing, int num_elements, double crater_diameter, double ejrim, double ejecta_truncation, double *ejecta_thickness);